﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace SetCBLUserImage
{
    public class CalculatorProgram
    {
        public CalculatorType Type { get; private set; }
        public byte DataTypeID { get; private set; }
        public string Name { get; private set; }
        public byte[] Data { get; private set; }

        public enum CalculatorType
        {
            TI73,
            TI82,
            TI83,
            TI83Plus,
            TI86,
            TI89,
            TI92,
            TI92Plus
        }

        private CalculatorProgram()
        {
            //Do nothing
        }

        public static CalculatorProgram Load(string fileName, bool ignoreChecksums)
        {
            var ret = new CalculatorProgram();
            var file = new FileStream(fileName, FileMode.Open);

            //Read header
            var header = new byte[8];
            file.Read(header, 0, header.Length);
            ret.Type = _GetCalculatorType(ASCIIEncoding.ASCII.GetString(header, 0, 8));

            int checksum;
            switch (ret.Type)
            {
                case CalculatorType.TI73:
                case CalculatorType.TI82:
                case CalculatorType.TI83:
                case CalculatorType.TI83Plus:
                case CalculatorType.TI86:
                    //Skip to variable entry
                    file.Seek(0x37, SeekOrigin.Begin);
                    var size = new byte[2];
                    file.Read(size, 0, size.Length);

                    //Read the variable size and type
                    var varSize = new byte[2];
                    file.Read(varSize, 0, varSize.Length);
                    ret.DataTypeID = (byte)file.ReadByte();

                    //The TI-86 has a variable name length byte
                    if (ret.Type == CalculatorType.TI86)
                    {
                        var nameLength = (byte)file.ReadByte();
                    }

                    //Read the variable name
                    var bytes = new byte[8];
                    file.Read(bytes, 0, bytes.Length);
                    ret.Name = ASCIIEncoding.ASCII.GetString(bytes).Trim(new char[] { '\0' }).Trim();

                    //Read the variable data
                    ret.Data = new byte[(varSize[0] | (varSize[1] << 8)) - 2];
                    file.Seek(0x37 + 6 + (size[0] | (size[1] << 8)), SeekOrigin.Begin);
                    file.Read(ret.Data, 0, ret.Data.Length);

                    //Read the checksum
                    var checksumBytes = new byte[2];
                    file.Read(checksumBytes, 0, checksumBytes.Length);
                    checksum = checksumBytes[0] | (checksumBytes[1] << 8);

                    //Validate it
                    int c = 0;
                    file.Seek(0x37, SeekOrigin.Begin);
                    var all = new byte[file.Length - 2 - 0x37];
                    file.Read(all, 0, all.Length);
                    foreach (var b in all)
                    {
                        c = (c + b) & 0xFFFF;
                    }

                    if (!ignoreChecksums && checksum != c)
                    {
                        throw new InvalidOperationException("Checksum is invalid");
                    }

                    break;
                case CalculatorType.TI89:
                case CalculatorType.TI92:
                case CalculatorType.TI92Plus:
                    //Skip to name
                    file.Seek(0x40, SeekOrigin.Begin);
                    var n = new byte[8];
                    file.Read(n, 0, n.Length);
                    ret.Name = ASCIIEncoding.ASCII.GetString(n).Trim(new char[] { '\0' }).Trim();

                    //Get data type
                    ret.DataTypeID = (byte)file.ReadByte();

                    //Skip to file size
                    var fileSize = new byte[4];
                    file.Seek(0x4C, SeekOrigin.Begin);
                    file.Read(fileSize, 0, fileSize.Length);

                    //Skip to file data and read it
                    var next = fileSize[0] | (fileSize[1] << 8) | (fileSize[2] << 16) | (fileSize[3] << 24);
                    next -= 2;
                    ret.Data = new byte[next - (file.Position + 2)];
                    file.Seek(0x52, SeekOrigin.Begin);
                    file.Read(ret.Data, 0, ret.Data.Length);

                    //Read the checksum
                    var checkBytes = new byte[2];
                    file.Read(checkBytes, 0, checkBytes.Length);
                    checksum = checkBytes[0] | (checkBytes[1] << 8);

                    //Validate it
                    int chk = 0;
                    foreach (var b in ret.Data)
                    {
                        chk = (chk + b) & 0xFFFF;
                    }

                    if (!ignoreChecksums && checksum != chk)
                    {
                        throw new InvalidOperationException("Checksum is invalid");
                    }

                    break;
                default:
                    throw new InvalidOperationException("Invalid type: " + ret.Type.ToString());
            }

            file.Close();

            return ret;
        }

        private static CalculatorType _GetCalculatorType(string header)
        {
            CalculatorType ret;

            switch (header)
            {
                case "**TI73**":
                    ret = CalculatorType.TI73;
                    break;
                case "**TI82**":
                    ret = CalculatorType.TI82;
                    break;
                case "**TI83**":
                    ret = CalculatorType.TI83;
                    break;
                case "**TI83F*":
                    ret = CalculatorType.TI83Plus;
                    break;
                case "**TI86**":
                    ret = CalculatorType.TI86;
                    break;
                case "**TI89**":
                    ret = CalculatorType.TI89;
                    break;
                case "**TI92**":
                    ret = CalculatorType.TI92;
                    break;
                case "**TI92P*":
                    ret = CalculatorType.TI92Plus;
                    break;
                default:
                    throw new InvalidOperationException("Invalid header: " + header);
            }

            return ret;
        }
    }
}
